"""Define the construction process.


"""


import sys
import signal
from copy import deepcopy
from contextlib import suppress
from queue import Empty
from multiprocessing import Queue

from cml.shared.logger import ConstructionInfo
from cml.domain.construction.construction_parallel import (
    BackgroundJobStatus,
    BackgroundConstructor)


def handle_background_constructor_failure(signum, stack):
    sys.stdout.write("An exception occured in the background constructor. "
                     "Tried to label a learnblock. "
                     "Refer to stderr to identify the exception cause.")
    sys.exit(-777)


signal.signal(signal.SIGALRM, handle_background_constructor_failure)


class AbstractConstructor:

    def __init__(self, mode, ml_models, settings):
        """

        Args:
            mode (str):
            ml_models (list):
            settings (ConstructionSettings):

        """
        self.mode = mode
        self.ml_models = ml_models
        self.settings = settings

    def construct(self, *args, **kwargs):
        raise NotImplementedError

    def to_dict(self):
        return {
            "mode": self.mode,
            "ml_models": self.ml_models,
            "settings": self.settings
        }

    def __str__(self):
        return "\n".join([
                "{:20}: {}".format("Mode", str(self.mode)),
                "{:20}: {}".format("ML Models", [
                    str(m) for m in self.ml_models]),
                str(self.settings),
            ]
        )

    def __repr__(self):
        return ", ".join([
            "{}={}".format("mode", repr(self.mode)),
            "{}".format(repr(self.settings)),
            "{}={}".format("ml_models", repr(self.ml_models))
            ]
        )


class SerialConstructor(AbstractConstructor):

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()

    def construct(self, learnblock):
        """

        Args:
            learnblock (PandasBlock):

        Returns:

        """
        if learnblock.labeled():
            yield ConstructionInfo(ConstructionInfo.State.PASSED), learnblock

        else:
            for model in self.machine_learning_models():
                trained_model = model.train(learnblock)

                if not self.valid_trained(trained_model):
                    yield ConstructionInfo(
                            state=ConstructionInfo.State.DISCARDED,
                            subject=trained_model.abbreviation,
                            n_cluster=model.found_cluster), \
                          None

                else:
                    yield ConstructionInfo(
                            state=ConstructionInfo.State.CONSTRUCTED,
                            subject=trained_model.abbreviation,
                            n_cluster=model.found_cluster), \
                        self.label_block(learnblock, trained_model)

    def machine_learning_models(self):
        for ml_model in self.ml_models:
            if ml_model.cluster_specific:
                for n_cluster in range(2, self.settings.max_categories+1):
                    ml_model.cluster = n_cluster
                    yield ml_model
            else:
                yield ml_model

    def valid_trained(self, model):
        """

        Args:
            model (MachineLearningModel):

        Returns:

        """
        if model.cluster_specific:
            if model.found_cluster != model.cluster:
                return False
        else:
            if model.found_cluster <= 1:
                return False

        for cluster, size in model.cluster_sizes.items():
            if size < self.settings.min_category_size:
                return False

        return True

    def label_block(self, block, trained_model):
        """

        Args:
            block (PandasBlock):
            trained_model (MachineLearningModel):

        Returns:

        """
        copied_block = deepcopy(block)
        labels = trained_model.get_labels()
        copied_block.set_column("Z", labels)
        copied_block.n_cluster = trained_model.found_cluster
        copied_block.origin = block.origin
        subject = "{}{:02}".format(trained_model.abbreviation,
                                   trained_model.found_cluster)
        copied_block.subject = subject
        return copied_block


class ParallelConstructor(AbstractConstructor):

    def __init__(self, mode, ml_models, settings):
        """

        Args:
            mode (str):
            ml_models (list):
            settings (ConstructionSettings):

        """
        super().__init__(mode, ml_models, settings)
        self.__process = None
        self.__input_queue = None
        self.__output_queue = None
        self.__control_queue = None

    def __str__(self):
        return super().__str__()

    def __repr__(self):
        return super().__repr__()

    def construct(self, learnblock):
        """

        Args:
            learnblock (PandasBlock):

        Returns:

        """
        self.prepare_background_process()
        self.__input_queue.put(learnblock)
        while 1:
            with suppress(Empty):
                response = self.__output_queue.get(timeout=0)
                if response == BackgroundJobStatus.DONE:
                    return
                else:
                    yield response

    def prepare_background_process(self):
        """

        Returns:

        """
        if not self.__process or not self.__process.is_alive():
            self.__input_queue = Queue()
            self.__output_queue = Queue()
            self.__control_queue = Queue()
            background = BackgroundConstructor(
                self.__input_queue,
                self.__output_queue,
                SerialConstructor(self.mode, self.ml_models, self.settings)
            )
            background.name = "BackgroundConstructor"
            background.daemon = True
            background.start()
            self.__process = background


class Constructor:

    def __init__(self, mode, ml_models, settings):
        """

        Args:
            mode (str):
            ml_models (list):
            settings (ConstructionSettings):

        """
        self.state = SerialConstructor(mode, ml_models, settings)

    def __str__(self):
        return str(self.state)

    def __repr__(self):
        return repr(self.state)

    def construct(self, learnblock):
        """

        Args:
            learnblock (PandasBlock):

        Returns:

        """
        for log_info, block in self.state.construct(learnblock):
            yield log_info, block

    def transit(self, parallel):
        """

        Args:
            parallel (bool):

        Returns:

        """
        if parallel:
            self.state = ParallelConstructor(**self.state.to_dict())
        else:
            self.state = SerialConstructor(**self.state.to_dict())

    @property
    def mode(self):
        return self.state.mode

    @mode.setter
    def mode(self, value):
        self.state.mode = value

    @property
    def ml_models(self):
        return self.state.ml_models

    @ml_models.setter
    def ml_models(self, value):
        self.state.ml_models = value

    @property
    def max_target_error(self):
        return self.state.settings.max_target_error

    @max_target_error.setter
    def max_target_error(self, value):
        self.state.settings.max_target_error = value

    @property
    def max_model_targets(self):
        return self.state.settings.max_model_targets

    @max_model_targets.setter
    def max_model_targets(self, value):
        self.state.settings.max_model_targets = value

    @property
    def min_category_size(self):
        return self.state.settings.min_category_size

    @min_category_size.setter
    def min_category_size(self, value):
        self.state.settings.min_category_size = value

    @property
    def max_categories(self):
        return self.state.settings.max_categories

    @max_categories.setter
    def max_categories(self, value):
        self.state.settings.max_categories = value
